@@ -2,12 +2,12 @@
 /* NAME
 /*	bounce 8
 /* SUMMARY
-/*	Postfix message bounce or defer daemon
+/*	Postfix delivery status reports
 /* SYNOPSIS
 /*	\fBbounce\fR [generic Postfix daemon options]
 /* DESCRIPTION
 /*	The \fBbounce\fR(8) daemon maintains per-message log files with
-/*	non-delivery status information. Each log file is named after the
+/*	delivery status information. Each log file is named after the
 /*	queue file that it corresponds to, and is kept in a queue subdirectory
 /*	named after the service name in the \fBmaster.cf\fR file (either
 /*	\fBbounce\fR, \fBdefer\fR or \fBtrace\fR).
@@ -19,8 +19,9 @@
 /*	Append a recipient (non-)delivery status record to a per-message
 /*	log file.
 /* .IP \(bu
-/*	Enqueue a bounce message, with a copy of a per-message log file
-/*	and of the corresponding message. When the bounce message is
+/*	Enqueue a delivery status notification message, with a copy
+/*	of a per-message log file and of the corresponding message.
+/*	When the delivery status notification message is
 /*	enqueued successfully, the per-message log file is deleted.
 /* .PP
 /*	The software does a best notification effort. A non-delivery
@@ -33,12 +34,18 @@
 /*	themselves, and that depend on retry logic in their own client.
 /* STANDARDS
 /*	RFC 822 (ARPA Internet Text Messages)
-/*	RFC 2822 (ARPA Internet Text Messages)
+/*	RFC 2045 (Format of Internet Message Bodies)
+/*	RFC 2822 (Internet Message Format)
 /*	RFC 3462 (Delivery Status Notifications)
 /*	RFC 3464 (Delivery Status Notifications)
-/*	RFC 2045 (Format of Internet Message Bodies)
+/*	RFC 3834 (Auto-Submitted: message header)
+/*	RFC 5322 (Internet Message Format)
+/*	RFC 6531 (Internationalized SMTP)
+/*	RFC 6532 (Internationalized Message Format)
+/*	RFC 6533 (Internationalized Delivery Status Notifications)
 /* DIAGNOSTICS
-/*	Problems and transactions are logged to \fBsyslogd\fR(8).
+/*	Problems and transactions are logged to \fBsyslogd\fR(8)
+/*	or \fBpostlogd\fR(8).
 /* CONFIGURATION PARAMETERS
 /* .ad
 /* .fi
@@ -82,15 +89,19 @@
 /* .IP "\fBipc_timeout (3600s)\fR"
 /*	The time limit for sending or receiving information over an internal
 /*	communication channel.
+/* .IP "\fBinternal_mail_filter_classes (empty)\fR"
+/*	What categories of Postfix-generated mail are subject to
+/*	before-queue content inspection by non_smtpd_milters, header_checks
+/*	and body_checks.
 /* .IP "\fBmail_name (Postfix)\fR"
 /*	The mail system name that is displayed in Received: headers, in
 /*	the SMTP greeting banner, and in bounced mail.
 /* .IP "\fBmax_idle (100s)\fR"
-/*	The maximum amount of time that an idle Postfix daemon process
-/*	waits for the next service request before exiting.
+/*	The maximum amount of time that an idle Postfix daemon process waits
+/*	for an incoming connection before terminating voluntarily.
 /* .IP "\fBmax_use (100)\fR"
-/*	The maximal number of connection requests before a Postfix daemon
-/*	process terminates.
+/*	The maximal number of incoming connections that a Postfix daemon
+/*	process will service before terminating voluntarily.
 /* .IP "\fBnotify_classes (resource, software)\fR"
 /*	The list of error classes that are reported to the postmaster.
 /* .IP "\fBprocess_id (read-only)\fR"
@@ -101,9 +112,18 @@
 /*	The location of the Postfix top-level queue directory.
 /* .IP "\fBsyslog_facility (mail)\fR"
 /*	The syslog facility of Postfix logging.
-/* .IP "\fBsyslog_name (postfix)\fR"
-/*	The mail system name that is prepended to the process name in syslog
-/*	records, so that "smtpd" becomes, for example, "postfix/smtpd".
+/* .IP "\fBsyslog_name (see 'postconf -d' output)\fR"
+/*	A prefix that is prepended to the process name in syslog
+/*	records, so that, for example, "smtpd" becomes "prefix/smtpd".
+/* .PP
+/*	Available in Postfix 3.0 and later:
+/* .IP "\fBsmtputf8_autodetect_classes (sendmail, verify)\fR"
+/*	Detect that a message requires SMTPUTF8 support for the specified
+/*	mail origin classes.
+/* .PP
+/*	Available in Postfix 3.3 and later:
+/* .IP "\fBservice_name (read-only)\fR"
+/*	The master.cf service name of a Postfix daemon process.
 /* FILES
 /*	/var/spool/postfix/bounce/* non-delivery records
 /*	/var/spool/postfix/defer/* non-delivery records
@@ -114,6 +134,7 @@
 /*	postconf(5), configuration parameters
 /*	master(5), generic daemon options
 /*	master(8), process manager
+/*	postlogd(8), Postfix logging
 /*	syslogd(8), system logging
 /* LICENSE
 /* .ad
@@ -124,6 +145,11 @@
 /*	IBM T.J. Watson Research
 /*	P.O. Box 704
 /*	Yorktown Heights, NY 10598, USA
+/*
+/*	Wietse Venema
+/*	Google, Inc.
+/*	111 8th Avenue
+/*	New York, NY 10011, USA
 /*--*/
 
 /* System library. */
@@ -132,22 +158,20 @@
 #include <string.h>
 #include <stdlib.h>
 
-#ifdef STRCASECMP_IN_STRINGS_H
-#include <strings.h>
-#endif
-
 /* Utility library. */
 
 #include <msg.h>
 #include <vstring.h>
 #include <vstream.h>
 #include <stringops.h>
+#include <load_file.h>
 
 /* Global library. */
 
 #include <mail_proto.h>
 #include <mail_queue.h>
 #include <mail_params.h>
+#include <mail_version.h>
 #include <mail_conf.h>
 #include <bounce.h>
 #include <mail_addr.h>
@@ -160,7 +184,7 @@
 
 /* Application-specific. */
 
-#include "bounce_service.h"
+#include <bounce_service.h>
 
  /*
   * Tunables.
@@ -186,40 +210,59 @@ static VSTRING *dsn_envid;
 static VSTRING *verp_delims;
 static DSN_BUF *dsn_buf;
 
+ /*
+  * Templates.
+  */
+BOUNCE_TEMPLATES *bounce_templates;
+
 #define STR vstring_str
 
+#define VS_NEUTER(s) printable(vstring_str(s), '?')
+
 /* bounce_append_proto - bounce_append server protocol */
 
 static int bounce_append_proto(char *service_name, VSTREAM *client)
 {
-    char   *myname = "bounce_append_proto";
+    const char *myname = "bounce_append_proto";
     int     flags;
-    RECIPIENT_VAR rcpt;
-    DSN_VAR dsn;
 
     /*
      * Read and validate the client request.
      */
     if (mail_command_server(client,
-			    ATTR_TYPE_NUM, MAIL_ATTR_FLAGS, &flags,
-			    ATTR_TYPE_STR, MAIL_ATTR_QUEUEID, queue_id,
-			    ATTR_TYPE_FUNC, rcpb_scan, (void *) rcpt_buf,
-			    ATTR_TYPE_FUNC, dsb_scan, (void *) dsn_buf,
+			    RECV_ATTR_INT(MAIL_ATTR_FLAGS, &flags),
+			    RECV_ATTR_STR(MAIL_ATTR_QUEUEID, queue_id),
+			    RECV_ATTR_FUNC(rcpb_scan, (void *) rcpt_buf),
+			    RECV_ATTR_FUNC(dsb_scan, (void *) dsn_buf),
 			    ATTR_TYPE_END) != 4) {
 	msg_warn("malformed request");
 	return (-1);
     }
+
+    /*
+     * Sanitize input.
+     */
     if (mail_queue_id_ok(STR(queue_id)) == 0) {
 	msg_warn("malformed queue id: %s", printable(STR(queue_id), '?'));
 	return (-1);
     }
-    (void) RECIPIENT_FROM_RCPT_BUF(&rcpt, rcpt_buf);
-    (void) DSN_FROM_DSN_BUF(&dsn, dsn_buf);
+    VS_NEUTER(rcpt_buf->address);
+    VS_NEUTER(rcpt_buf->orig_addr);
+    VS_NEUTER(rcpt_buf->dsn_orcpt);
+    VS_NEUTER(dsn_buf->status);
+    VS_NEUTER(dsn_buf->action);
+    VS_NEUTER(dsn_buf->reason);
+    VS_NEUTER(dsn_buf->dtype);
+    VS_NEUTER(dsn_buf->dtext);
+    VS_NEUTER(dsn_buf->mtype);
+    VS_NEUTER(dsn_buf->mname);
+    (void) RECIPIENT_FROM_RCPT_BUF(rcpt_buf);
+    (void) DSN_FROM_DSN_BUF(dsn_buf);
 
     /*
-     * Beware: some dsn or rcpt fields may be null; access dsn_buf and
-     * rcpt_buf instead. See DSN_FROM_DSN_BUF(), RECIPIENT_FROM_RCPT_BUF(),
-     * and bounce_log(3).
+     * Beware: some DSN or RECIPIENT fields may be null; access dsn_buf and
+     * rcpt_buf buffers instead. See DSN_FROM_DSN_BUF() and
+     * RECIPIENT_FROM_RCPT_BUF().
      */
     if (msg_verbose)
 	msg_info("%s: flags=0x%x service=%s id=%s org_to=%s to=%s off=%ld dsn_org=%s, notif=0x%x stat=%s act=%s why=%s",
@@ -240,34 +283,41 @@ static int bounce_append_proto(char *service_name, VSTREAM *client)
      * Execute the request.
      */
     return (bounce_append_service(flags, service_name, STR(queue_id),
-				  &rcpt, &dsn));
+				  &rcpt_buf->rcpt, &dsn_buf->dsn));
 }
 
 /* bounce_notify_proto - bounce_notify server protocol */
 
 static int bounce_notify_proto(char *service_name, VSTREAM *client,
 			        int (*service) (int, char *, char *, char *,
-				               char *, char *, char *, int))
+				           char *, int, char *, char *, int,
+						        BOUNCE_TEMPLATES *))
 {
-    char   *myname = "bounce_notify_proto";
+    const char *myname = "bounce_notify_proto";
     int     flags;
+    int     smtputf8;
     int     dsn_ret;
 
     /*
      * Read and validate the client request.
      */
     if (mail_command_server(client,
-			    ATTR_TYPE_NUM, MAIL_ATTR_FLAGS, &flags,
-			    ATTR_TYPE_STR, MAIL_ATTR_QUEUE, queue_name,
-			    ATTR_TYPE_STR, MAIL_ATTR_QUEUEID, queue_id,
-			    ATTR_TYPE_STR, MAIL_ATTR_ENCODING, encoding,
-			    ATTR_TYPE_STR, MAIL_ATTR_SENDER, sender,
-			    ATTR_TYPE_STR, MAIL_ATTR_DSN_ENVID, dsn_envid,
-			    ATTR_TYPE_NUM, MAIL_ATTR_DSN_RET, &dsn_ret,
-			    ATTR_TYPE_END) != 7) {
+			    RECV_ATTR_INT(MAIL_ATTR_FLAGS, &flags),
+			    RECV_ATTR_STR(MAIL_ATTR_QUEUE, queue_name),
+			    RECV_ATTR_STR(MAIL_ATTR_QUEUEID, queue_id),
+			    RECV_ATTR_STR(MAIL_ATTR_ENCODING, encoding),
+			    RECV_ATTR_INT(MAIL_ATTR_SMTPUTF8, &smtputf8),
+			    RECV_ATTR_STR(MAIL_ATTR_SENDER, sender),
+			    RECV_ATTR_STR(MAIL_ATTR_DSN_ENVID, dsn_envid),
+			    RECV_ATTR_INT(MAIL_ATTR_DSN_RET, &dsn_ret),
+			    ATTR_TYPE_END) != 8) {
 	msg_warn("malformed request");
 	return (-1);
     }
+
+    /*
+     * Sanitize input.
+     */
     if (mail_queue_name_ok(STR(queue_name)) == 0) {
 	msg_warn("malformed queue name: %s", printable(STR(queue_name), '?'));
 	return (-1);
@@ -276,11 +326,14 @@ static int bounce_notify_proto(char *service_name, VSTREAM *client,
 	msg_warn("malformed queue id: %s", printable(STR(queue_id), '?'));
 	return (-1);
     }
-    printable(STR(dsn_envid), '?');
+    VS_NEUTER(encoding);
+    VS_NEUTER(sender);
+    VS_NEUTER(dsn_envid);
     if (msg_verbose)
-	msg_info("%s: flags=0x%x service=%s queue=%s id=%s encoding=%s sender=%s envid=%s ret=0x%x",
+	msg_info("%s: flags=0x%x service=%s queue=%s id=%s encoding=%s smtputf8=%d sender=%s envid=%s ret=0x%x",
 		 myname, flags, service_name, STR(queue_name), STR(queue_id),
-		 STR(encoding), STR(sender), STR(dsn_envid), dsn_ret);
+		 STR(encoding), smtputf8, STR(sender), STR(dsn_envid),
+		 dsn_ret);
 
     /*
      * On request by the client, set up a trap to delete the log file in case
@@ -293,34 +346,41 @@ static int bounce_notify_proto(char *service_name, VSTREAM *client,
      * Execute the request.
      */
     return (service(flags, service_name, STR(queue_name),
-		    STR(queue_id), STR(encoding),
-		    STR(sender), STR(dsn_envid), dsn_ret));
+		    STR(queue_id), STR(encoding), smtputf8,
+		    STR(sender), STR(dsn_envid), dsn_ret,
+		    bounce_templates));
 }
 
 /* bounce_verp_proto - bounce_notify server protocol, VERP style */
 
 static int bounce_verp_proto(char *service_name, VSTREAM *client)
 {
-    char   *myname = "bounce_verp_proto";
+    const char *myname = "bounce_verp_proto";
     int     flags;
+    int     smtputf8;
     int     dsn_ret;
 
     /*
      * Read and validate the client request.
      */
     if (mail_command_server(client,
-			    ATTR_TYPE_NUM, MAIL_ATTR_FLAGS, &flags,
-			    ATTR_TYPE_STR, MAIL_ATTR_QUEUE, queue_name,
-			    ATTR_TYPE_STR, MAIL_ATTR_QUEUEID, queue_id,
-			    ATTR_TYPE_STR, MAIL_ATTR_ENCODING, encoding,
-			    ATTR_TYPE_STR, MAIL_ATTR_SENDER, sender,
-			    ATTR_TYPE_STR, MAIL_ATTR_DSN_ENVID, dsn_envid,
-			    ATTR_TYPE_NUM, MAIL_ATTR_DSN_RET, &dsn_ret,
-			    ATTR_TYPE_STR, MAIL_ATTR_VERPDL, verp_delims,
-			    ATTR_TYPE_END) != 8) {
+			    RECV_ATTR_INT(MAIL_ATTR_FLAGS, &flags),
+			    RECV_ATTR_STR(MAIL_ATTR_QUEUE, queue_name),
+			    RECV_ATTR_STR(MAIL_ATTR_QUEUEID, queue_id),
+			    RECV_ATTR_STR(MAIL_ATTR_ENCODING, encoding),
+			    RECV_ATTR_INT(MAIL_ATTR_SMTPUTF8, &smtputf8),
+			    RECV_ATTR_STR(MAIL_ATTR_SENDER, sender),
+			    RECV_ATTR_STR(MAIL_ATTR_DSN_ENVID, dsn_envid),
+			    RECV_ATTR_INT(MAIL_ATTR_DSN_RET, &dsn_ret),
+			    RECV_ATTR_STR(MAIL_ATTR_VERPDL, verp_delims),
+			    ATTR_TYPE_END) != 9) {
 	msg_warn("malformed request");
 	return (-1);
     }
+
+    /*
+     * Sanitize input.
+     */
     if (mail_queue_name_ok(STR(queue_name)) == 0) {
 	msg_warn("malformed queue name: %s", printable(STR(queue_name), '?'));
 	return (-1);
@@ -329,16 +389,18 @@ static int bounce_verp_proto(char *service_name, VSTREAM *client)
 	msg_warn("malformed queue id: %s", printable(STR(queue_id), '?'));
 	return (-1);
     }
-    printable(STR(dsn_envid), '?');
+    VS_NEUTER(encoding);
+    VS_NEUTER(sender);
+    VS_NEUTER(dsn_envid);
+    VS_NEUTER(verp_delims);
     if (strlen(STR(verp_delims)) != 2) {
-	msg_warn("malformed verp delimiter string: %s",
-		 printable(STR(verp_delims), '?'));
+	msg_warn("malformed verp delimiter string: %s", STR(verp_delims));
 	return (-1);
     }
     if (msg_verbose)
-	msg_info("%s: flags=0x%x service=%s queue=%s id=%s encoding=%s sender=%s envid=%s ret=0x%x delim=%s",
+	msg_info("%s: flags=0x%x service=%s queue=%s id=%s encoding=%s smtputf8=%d sender=%s envid=%s ret=0x%x delim=%s",
 		 myname, flags, service_name, STR(queue_name),
-		 STR(queue_id), STR(encoding), STR(sender),
+		 STR(queue_id), STR(encoding), smtputf8, STR(sender),
 		 STR(dsn_envid), dsn_ret, STR(verp_delims));
 
     /*
@@ -352,45 +414,51 @@ static int bounce_verp_proto(char *service_name, VSTREAM *client)
      * Execute the request. Fall back to traditional notification if a bounce
      * was returned as undeliverable, because we don't want to VERPify those.
      */
-    if (!*STR(sender) || !strcasecmp(STR(sender), mail_addr_double_bounce())) {
+    if (!*STR(sender) || !strcasecmp_utf8(STR(sender),
+					  mail_addr_double_bounce())) {
 	msg_warn("request to send VERP-style notification of bounced mail");
 	return (bounce_notify_service(flags, service_name, STR(queue_name),
-				      STR(queue_id), STR(encoding),
-				      STR(sender), STR(dsn_envid), dsn_ret));
+				      STR(queue_id), STR(encoding), smtputf8,
+				      STR(sender), STR(dsn_envid), dsn_ret,
+				      bounce_templates));
     } else
 	return (bounce_notify_verp(flags, service_name, STR(queue_name),
-				   STR(queue_id), STR(encoding),
+				   STR(queue_id), STR(encoding), smtputf8,
 				   STR(sender), STR(dsn_envid), dsn_ret,
-				   STR(verp_delims)));
+				   STR(verp_delims), bounce_templates));
 }
 
 /* bounce_one_proto - bounce_one server protocol */
 
 static int bounce_one_proto(char *service_name, VSTREAM *client)
 {
-    char   *myname = "bounce_one_proto";
+    const char *myname = "bounce_one_proto";
     int     flags;
+    int     smtputf8;
     int     dsn_ret;
-    RECIPIENT rcpt;
-    DSN     dsn;
 
     /*
      * Read and validate the client request.
      */
     if (mail_command_server(client,
-			    ATTR_TYPE_NUM, MAIL_ATTR_FLAGS, &flags,
-			    ATTR_TYPE_STR, MAIL_ATTR_QUEUE, queue_name,
-			    ATTR_TYPE_STR, MAIL_ATTR_QUEUEID, queue_id,
-			    ATTR_TYPE_STR, MAIL_ATTR_ENCODING, encoding,
-			    ATTR_TYPE_STR, MAIL_ATTR_SENDER, sender,
-			    ATTR_TYPE_STR, MAIL_ATTR_DSN_ENVID, dsn_envid,
-			    ATTR_TYPE_NUM, MAIL_ATTR_DSN_RET, &dsn_ret,
-			    ATTR_TYPE_FUNC, rcpb_scan, (void *) rcpt_buf,
-			    ATTR_TYPE_FUNC, dsb_scan, (void *) dsn_buf,
-			    ATTR_TYPE_END) != 9) {
+			    RECV_ATTR_INT(MAIL_ATTR_FLAGS, &flags),
+			    RECV_ATTR_STR(MAIL_ATTR_QUEUE, queue_name),
+			    RECV_ATTR_STR(MAIL_ATTR_QUEUEID, queue_id),
+			    RECV_ATTR_STR(MAIL_ATTR_ENCODING, encoding),
+			    RECV_ATTR_INT(MAIL_ATTR_SMTPUTF8, &smtputf8),
+			    RECV_ATTR_STR(MAIL_ATTR_SENDER, sender),
+			    RECV_ATTR_STR(MAIL_ATTR_DSN_ENVID, dsn_envid),
+			    RECV_ATTR_INT(MAIL_ATTR_DSN_RET, &dsn_ret),
+			    RECV_ATTR_FUNC(rcpb_scan, (void *) rcpt_buf),
+			    RECV_ATTR_FUNC(dsb_scan, (void *) dsn_buf),
+			    ATTR_TYPE_END) != 10) {
 	msg_warn("malformed request");
 	return (-1);
     }
+
+    /*
+     * Sanitize input.
+     */
     if (strcmp(service_name, MAIL_SERVICE_BOUNCE) != 0) {
 	msg_warn("wrong service name \"%s\" for one-recipient bouncing",
 		 service_name);
@@ -404,20 +472,32 @@ static int bounce_one_proto(char *service_name, VSTREAM *client)
 	msg_warn("malformed queue id: %s", printable(STR(queue_id), '?'));
 	return (-1);
     }
-    printable(STR(dsn_envid), '?');
-    (void) RECIPIENT_FROM_RCPT_BUF(&rcpt, rcpt_buf);
-    (void) DSN_FROM_DSN_BUF(&dsn, dsn_buf);
+    VS_NEUTER(encoding);
+    VS_NEUTER(sender);
+    VS_NEUTER(dsn_envid);
+    VS_NEUTER(rcpt_buf->address);
+    VS_NEUTER(rcpt_buf->orig_addr);
+    VS_NEUTER(rcpt_buf->dsn_orcpt);
+    VS_NEUTER(dsn_buf->status);
+    VS_NEUTER(dsn_buf->action);
+    VS_NEUTER(dsn_buf->reason);
+    VS_NEUTER(dsn_buf->dtype);
+    VS_NEUTER(dsn_buf->dtext);
+    VS_NEUTER(dsn_buf->mtype);
+    VS_NEUTER(dsn_buf->mname);
+    (void) RECIPIENT_FROM_RCPT_BUF(rcpt_buf);
+    (void) DSN_FROM_DSN_BUF(dsn_buf);
 
     /*
-     * Beware: some dsn or rcpt fields may be null; access dsn_buf and
-     * rcpt_buf instead. See DSN_FROM_DSN_BUF(), RECIPIENT_FROM_RCPT_BUF(),
-     * and bounce_log(3).
+     * Beware: some DSN or RECIPIENT fields may be null; access dsn_buf and
+     * rcpt_buf buffers instead. See DSN_FROM_DSN_BUF() and
+     * RECIPIENT_FROM_RCPT_BUF().
      */
     if (msg_verbose)
-	msg_info("%s: flags=0x%x queue=%s id=%s encoding=%s sender=%s envid=%s dsn_ret=0x%x orig_to=%s to=%s off=%ld dsn_orig=%s notif=0x%x stat=%s act=%s why=%s",
+	msg_info("%s: flags=0x%x queue=%s id=%s encoding=%s smtputf8=%d sender=%s envid=%s dsn_ret=0x%x orig_to=%s to=%s off=%ld dsn_orig=%s notif=0x%x stat=%s act=%s why=%s",
 		 myname, flags, STR(queue_name), STR(queue_id),
-		 STR(encoding), STR(sender), STR(dsn_envid), dsn_ret,
-		 STR(rcpt_buf->orig_addr), STR(rcpt_buf->address),
+		 STR(encoding), smtputf8, STR(sender), STR(dsn_envid),
+		 dsn_ret, STR(rcpt_buf->orig_addr), STR(rcpt_buf->address),
 		 rcpt_buf->offset, STR(rcpt_buf->dsn_orcpt),
 		 rcpt_buf->dsn_notify, STR(dsn_buf->status),
 		 STR(dsn_buf->action), STR(dsn_buf->reason));
@@ -426,8 +506,9 @@ static int bounce_one_proto(char *service_name, VSTREAM *client)
      * Execute the request.
      */
     return (bounce_one_service(flags, STR(queue_name), STR(queue_id),
-			       STR(encoding), STR(sender), STR(dsn_envid),
-			       dsn_ret, &rcpt, &dsn));
+			       STR(encoding), smtputf8, STR(sender),
+			       STR(dsn_envid), dsn_ret, rcpt_buf,
+			       dsn_buf, bounce_templates));
 }
 
 /* bounce_service - parse bounce command type and delegate */
@@ -451,7 +532,7 @@ static void bounce_service(VSTREAM *client, char *service_name, char **argv)
      * request-specific protocol routines take care of the remainder.
      */
     if (attr_scan(client, ATTR_FLAG_STRICT | ATTR_FLAG_MORE,
-		  ATTR_TYPE_NUM, MAIL_ATTR_NREQ, &command, 0) != 1) {
+		  RECV_ATTR_INT(MAIL_ATTR_NREQ, &command), 0) != 1) {
 	msg_warn("malformed request");
 	status = -1;
     } else if (command == BOUNCE_CMD_VERP) {
@@ -479,7 +560,7 @@ static void bounce_service(VSTREAM *client, char *service_name, char **argv)
      * client.
      */
     attr_print(client, ATTR_FLAG_NONE,
-	       ATTR_TYPE_NUM, MAIL_ATTR_STATUS, status,
+	       SEND_ATTR_INT(MAIL_ATTR_STATUS, status),
 	       ATTR_TYPE_END);
     vstream_fflush(client);
 
@@ -495,39 +576,53 @@ static void bounce_service(VSTREAM *client, char *service_name, char **argv)
     }
 }
 
+static void load_helper(VSTREAM *stream, void *context)
+{
+    BOUNCE_TEMPLATES *templates = (BOUNCE_TEMPLATES *) context;
+
+    bounce_templates_load(stream, templates);
+}
+
 /* pre_jail_init - pre-jail initialization */
 
-static void pre_jail_init(char *service_name, char **unused_argv)
+static void pre_jail_init(char *unused_name, char **unused_argv)
 {
 
     /*
-     * Load the alternate message files (if specified) before entering the ch
-     * root jail.
+     * Bundle up a bunch of bounce template information.
      */
-    if (*var_bounce_tmpl)
-	bounce_template_load(var_bounce_tmpl);
+    bounce_templates = bounce_templates_create();
 
     /*
-     * Special case: dump bounce templates. This is not part of the
-     * master(5) public interface.
+     * Load the alternate message files (if specified) before entering the
+     * chroot jail.
      */
-    if (strcmp(service_name, "dump_templates") == 0) {
-	bounce_template_dump_all(VSTREAM_OUT);
-	vstream_fflush(VSTREAM_OUT);
-	exit(0);
-    }
-    if (strcmp(service_name, "expand_templates") == 0) {
-	bounce_template_expand_all(VSTREAM_OUT);
-	vstream_fflush(VSTREAM_OUT);
-	exit(0);
-    }
+    if (*var_bounce_tmpl)
+	load_file(var_bounce_tmpl, load_helper, (void *) bounce_templates);
 }
 
 /* post_jail_init - initialize after entering chroot jail */
 
-static void post_jail_init(char *unused_name, char **unused_argv)
+static void post_jail_init(char *service_name, char **unused_argv)
 {
 
+    /*
+     * Special case: dump bounce templates. This is not part of the master(5)
+     * public interface. This internal interface is used by the postconf
+     * command. It was implemented before bounce templates were isolated into
+     * modules that could have been called directly.
+     */
+    if (strcmp(service_name, "dump_templates") == 0) {
+	bounce_templates_dump(VSTREAM_OUT, bounce_templates);
+	vstream_fflush(VSTREAM_OUT);
+	exit(0);
+    }
+    if (strcmp(service_name, "expand_templates") == 0) {
+	bounce_templates_expand(VSTREAM_OUT, bounce_templates);
+	vstream_fflush(VSTREAM_OUT);
+	exit(0);
+    }
+
     /*
      * Initialize. We're single threaded so we can reuse some memory upon
      * successive requests.
@@ -542,20 +637,22 @@ static void post_jail_init(char *unused_name, char **unused_argv)
     dsn_buf = dsb_create();
 }
 
+MAIL_VERSION_STAMP_DECLARE;
+
 /* main - the main program */
 
 int     main(int argc, char **argv)
 {
-    static CONFIG_INT_TABLE int_table[] = {
+    static const CONFIG_INT_TABLE int_table[] = {
 	VAR_BOUNCE_LIMIT, DEF_BOUNCE_LIMIT, &var_bounce_limit, 1, 0,
 	0,
     };
-    static CONFIG_TIME_TABLE time_table[] = {
+    static const CONFIG_TIME_TABLE time_table[] = {
 	VAR_MAX_QUEUE_TIME, DEF_MAX_QUEUE_TIME, &var_max_queue_time, 0, 8640000,
 	VAR_DELAY_WARN_TIME, DEF_DELAY_WARN_TIME, &var_delay_warn_time, 0, 0,
 	0,
     };
-    static CONFIG_STR_TABLE str_table[] = {
+    static const CONFIG_STR_TABLE str_table[] = {
 	VAR_NOTIFY_CLASSES, DEF_NOTIFY_CLASSES, &var_notify_classes, 0, 0,
 	VAR_BOUNCE_RCPT, DEF_BOUNCE_RCPT, &var_bounce_rcpt, 1, 0,
 	VAR_2BOUNCE_RCPT, DEF_2BOUNCE_RCPT, &var_2bounce_rcpt, 1, 0,
@@ -564,15 +661,20 @@ int     main(int argc, char **argv)
 	0,
     };
 
+    /*
+     * Fingerprint executables and core dumps.
+     */
+    MAIL_VERSION_STAMP_ALLOCATE;
+
     /*
      * Pass control to the single-threaded service skeleton.
      */
     single_server_main(argc, argv, bounce_service,
-		       MAIL_SERVER_INT_TABLE, int_table,
-		       MAIL_SERVER_STR_TABLE, str_table,
-		       MAIL_SERVER_TIME_TABLE, time_table,
-		       MAIL_SERVER_PRE_INIT, pre_jail_init,
-		       MAIL_SERVER_POST_INIT, post_jail_init,
-		       MAIL_SERVER_UNLIMITED,
+		       CA_MAIL_SERVER_INT_TABLE(int_table),
+		       CA_MAIL_SERVER_STR_TABLE(str_table),
+		       CA_MAIL_SERVER_TIME_TABLE(time_table),
+		       CA_MAIL_SERVER_PRE_INIT(pre_jail_init),
+		       CA_MAIL_SERVER_POST_INIT(post_jail_init),
+		       CA_MAIL_SERVER_UNLIMITED,
 		       0);
 }
